@@ -28,14 +28,17 @@ require 'yaml'
 
 module IcdGenerator
   $api_entries = {}
+  $key_entries = {}
   $api_entries_array = []
   $cl_objects = ["platform_id", "device_id", "context", "command_queue", "mem", "program", "kernel", "event", "sampler"]
-  $know_entries = { 1 => "clGetPlatformInfo", 0 => "clGetPlatformIDs" }
-  $use_name_in_test = { 1 => "clGetPlatformInfo", 0 => "clGetPlatformIDs" }
+  $know_entries = { 1 => "clGetPlatformInfo", 0 => "clGetPlatformIDs", 2 => "clGetDeviceIDs" }
+  $use_name_in_test = { 1 => "clGetPlatformInfo", 0 => "clGetPlatformIDs", 2 => "clGetDeviceIDs" }
   # do not call these functions when trying to discover the mapping
-  $forbidden_funcs = ["clGetExtensionFunctionAddress", "clGetPlatformIDs",
-    "clGetPlatformInfo", "clGetGLContextInfoKHR", "clUnloadCompiler",
-    "clSetCommandQueueProperty"]
+  $forbidden_funcs = [
+    "clUnloadCompiler", # No parameters so no way to forward to an ICD, OpenCL 1.0
+    "clGetPlatformIDs", # Implemented directly within the ICD Loader, not forwarded to ICD
+  ]
+  # windows function to ignore when loading the database
   $windows_funcs = ["clGetDeviceIDsFromD3D10KHR", "clCreateFromD3D10BufferKHR",
     "clCreateFromD3D10Texture2DKHR", "clCreateFromD3D10Texture3DKHR",
     "clEnqueueAcquireD3D10ObjectsKHR", "clEnqueueReleaseD3D10ObjectsKHR",
@@ -141,6 +144,7 @@ EOF
       version = value.match(/SUFFIX__VERSION_(\d_\d)/m)[1]
       $versions_entries[version].push(entry_name)
       $known_entries[key] = entry_name
+      $key_entries[entry_name] = key
       $api_entries[entry_name] = value
     }
     $api_entries_array = []
@@ -223,13 +227,17 @@ EOF
     run_dummy_icd += "#  pragma GCC diagnostic ignored \"-Wcpp\"\n"
     run_dummy_icd += "#  define CL_USE_DEPRECATED_OPENCL_1_0_APIS\n"
     run_dummy_icd += "#  define CL_USE_DEPRECATED_OPENCL_1_1_APIS\n"
+    run_dummy_icd += "#  define CL_USE_DEPRECATED_OPENCL_1_2_APIS\n"
     run_dummy_icd += "#  include <CL/opencl.h>\n"
     run_dummy_icd += self.include_headers
     run_dummy_icd += "#pragma GCC diagnostic pop\n"
     run_dummy_icd += "\n\n"
-    run_dummy_icd += "typedef CL_API_ENTRY cl_int (CL_API_CALL* oclFuncPtr_fn)(cl_platform_id platform);\n\n"
+    $api_entries.each_key { |func_name|
+       next if $forbidden_funcs.include?(func_name)
+       run_dummy_icd += $api_entries[func_name]+";\n"
+    }
+    run_dummy_icd += "\n\n"
     run_dummy_icd += "void call_all_OpenCL_functions(cl_platform_id chosen_platform) {\n"
-    run_dummy_icd += "  oclFuncPtr_fn oclFuncPtr;\n"
     run_dummy_icd += "  cl_context_properties properties[] = { CL_CONTEXT_PLATFORM, (cl_context_properties)chosen_platform, 0 };\n"
     $api_entries.each_key { |func_name|
        next if $forbidden_funcs.include?(func_name)
@@ -241,13 +249,32 @@ EOF
          run_dummy_icd += "  #{func_name}(properties,CL_DEVICE_TYPE_CPU,NULL,NULL,NULL);\n"
        elsif func_name == "clWaitForEvents" then
          run_dummy_icd += "  #{func_name}(1,(cl_event*)&chosen_platform);\n"
-       elsif func_name == "clGetExtensionFunctionAddressForPlatform" then
-         run_dummy_icd += "  #{func_name}((cl_platform_id)chosen_platform, \"clIcdGetPlatformIDsKHR\");\n"
+       elsif func_name == "clGetExtensionFunctionAddress" then
+         run_dummy_icd += "  #{func_name}(\"extLIG\");\n"
+       elsif func_name == "clUnloadCompiler" then
+         run_dummy_icd += "  #{func_name}();\n"
        else
-         run_dummy_icd += "  oclFuncPtr = (oclFuncPtr_fn)" + func_name + ";\n"
-         run_dummy_icd += "  oclFuncPtr(chosen_platform);\n"
+         params = $api_entries[func_name].gsub(/[[:space:]]+/, ' ')
+         params.gsub!(/[\/][*](([^*]*)[*][^\/])*[^*]*[*][\/]/,'')
+         params.gsub!(/^[^(]*[(] *(.*[^ ]) *[)][^)]*$/, '\1')
+         params = params.gsub(/[^,(]+([(][^)]*[)])*[^,]*/) { |p|
+           p.gsub!("[]", "*")
+           p.gsub!("user_func", "")
+           if false && p.match(/[*]/) then
+             "NULL"
+           else
+             "(#{p})0"
+           end
+         }
+         if func_name != "clGetPlatformIDs" then
+           params.gsub!(/^([^0]*)0/, '\1chosen_platform')
+         end
+         run_dummy_icd += "  #{func_name}(#{params});\n"
        end
-       run_dummy_icd += "  printf(\"%s\\n\", \"#{func_name}\");"
+       run_dummy_icd += "  printf(\"%s\\n\", \"#{func_name}\");\n"
+       run_dummy_icd += "#ifdef OCL_ICD_PRINT_EXPECTED\n"
+       run_dummy_icd += "  printf(\"#{$key_entries[func_name]}  : %s (expected)\\n\", \"#{func_name}\");\n"
+       run_dummy_icd += "#endif\n"
        run_dummy_icd += "  fflush(NULL);\n"
     }
     run_dummy_icd += "  return;\n}\n"
@@ -261,22 +288,47 @@ EOF
 #include <stdio.h>
 #include <dlfcn.h>
 
-#define F(f) \\
-__attribute__((weak)) int f (void* arg, void* arg2) { \\
-  void (* p)(void*, void*)=NULL; \\
-  p=dlsym(RTLD_NEXT, #f); \\
-  if (p) { \\
-    (*p)(arg, arg2); \\
-  } else { \\
-    printf("-1 : "); \\
-  } \\
-  return 0; \\
-} 
-
 EOF
-    $api_entries.each_key { |func_name|
-       next if $noweak_funcs.include?(func_name)
-       run_dummy_icd_weak += "F(#{func_name})\n"
+    run_dummy_icd_weak += "#pragma GCC diagnostic push\n"
+    run_dummy_icd_weak += "#  pragma GCC diagnostic ignored \"-Wcpp\"\n"
+    run_dummy_icd_weak += "#  define CL_USE_DEPRECATED_OPENCL_1_0_APIS\n"
+    run_dummy_icd_weak += "#  define CL_USE_DEPRECATED_OPENCL_1_1_APIS\n"
+    run_dummy_icd_weak += "#  define CL_USE_DEPRECATED_OPENCL_1_2_APIS\n"
+    run_dummy_icd_weak += "#  include <CL/opencl.h>\n"
+    run_dummy_icd_weak += self.include_headers
+    run_dummy_icd_weak += "#pragma GCC diagnostic pop\n"
+
+    $api_entries.each { |func_name, entry|
+      next if $noweak_funcs.include?(func_name)
+      clean_entry = entry.sub(/(.*\)).*/m,'\1').gsub("/*","").gsub("*/","").gsub("\r","") + "{\n"
+      return_type = entry.match(/CL_API_ENTRY (.*) CL_API_CALL/)[1]
+      parameters = clean_entry.match(/\(.*\)/m)[0][1..-2]
+      parameters.gsub!(/\[.*?\]/,"")
+      parameters.sub!(/\(.*?\*\s*(.*?)\)\s*\(.*?\)/m,'\1')
+      run_dummy_icd_weak += clean_entry.gsub(/\*\[.*?\]/,"*  ").gsub(/\[.+?\]/,"")
+      first_parameter = parameters.match(/.*?\,/m)
+      if not first_parameter then
+        first_parameter =  parameters.match(/.*/m)[0]
+      else
+        first_parameter = first_parameter[0][0..-2]
+      end
+      fps = first_parameter.split
+      run_dummy_icd_weak += "  void (*p)()=dlsym(RTLD_NEXT, \"#{func_name}\");\n"
+      ps = parameters.split(",")
+      ps = ps.collect { |p|
+        p = p.split
+        p = p[-1].gsub("*","")
+      }
+       
+      run_dummy_icd_weak += "  if(p) {\n"
+      run_dummy_icd_weak += "    return((*(typeof(#{func_name})*)p)("
+      run_dummy_icd_weak += ps.join(", ")
+      run_dummy_icd_weak += "));\n"
+      run_dummy_icd_weak += "  } else {\n"
+      run_dummy_icd_weak += "    printf(\"-1 : \");\n"
+      run_dummy_icd_weak += "    return (#{return_type})0;\n"
+      run_dummy_icd_weak += "  }\n"
+      run_dummy_icd_weak += "}\n\n"
     }
     return run_dummy_icd_weak
   end
@@ -362,15 +414,18 @@ typedef __typeof__(clGetExtensionFunctionAddress) *clGetExtensionFunctionAddress
 extern const struct func_desc function_description[];
 struct vendor_icd {
   cl_uint	num_platforms;
-  cl_uint	first_platform;
   void *	dl_handle;
   clGetExtensionFunctionAddress_fn ext_fn_ptr;
 };
+
 struct platform_icd {
-  char *	 extension_suffix;
-  char *	 version;
+  char *	extension_suffix;
+  char *	version;
   struct vendor_icd *vicd;
   cl_platform_id pid;
+  cl_uint	ngpus; /* number of GPU devices */
+  cl_uint	ncpus; /* number of CPU devices */
+  cl_uint	ndevs; /* total number of devices, of all types */
 };
 
 EOF
@@ -574,12 +629,13 @@ EOF
       error_handler.call
       ocl_icd_loader_gen_source += "  }\n"
       if return_type != "void" then
-        ocl_icd_loader_gen_source += "  RETURN(((struct _#{fps[0]} *)#{fps[1]})->dispatch->#{func_name}("
-        ocl_icd_loader_gen_source += ps.join(", ")
-        ocl_icd_loader_gen_source += "));\n"
+        return_debug="RETURN"
       else
-        ocl_icd_loader_gen_source += "  return;"
+        return_debug="return"
       end
+      ocl_icd_loader_gen_source += "  #{return_debug}(((struct _#{fps[0]} *)#{fps[1]})->dispatch->#{func_name}("
+      ocl_icd_loader_gen_source += ps.join(", ")
+      ocl_icd_loader_gen_source += "));\n"
       ocl_icd_loader_gen_source += "}\n\n"
     }
     ocl_icd_loader_gen_source += "#pragma GCC visibility push(hidden)\n\n"
